#lang racket


;;        .__                   __                                                  __   
;;   ____ |  |__   ____   ____ |  | __           ____ ___  _________   ____   _____/  |_ 
;; _/ ___\|  |  \_/ __ \_/ ___\|  |/ /  ______ _/ __ \\  \/  /\____ \_/ __ \_/ ___\   __\
;; \  \___|   Y  \  ___/\  \___|    <  /_____/ \  ___/ >    < |  |_> >  ___/\  \___|  |  
;;  \___  >___|  /\___  >\___  >__|_ \          \___  >__/\_ \|   __/ \___  >\___  >__|  
;;      \/     \/     \/     \/     \/              \/      \/|__|        \/     \/      
;; _________        .__  .__                 __                                          
;; \_   ___ \  ____ |  | |  |   ____   _____/  |_  ___________                           
;; /    \  \/ /  _ \|  | |  | _/ __ \_/ ___\   __\/  _ \_  __ \                          
;; \     \___(  <_> )  |_|  |_\  ___/\  \___|  | (  <_> )  | \/                          
;;  \______  /\____/|____/____/\___  >\___  >__|  \____/|__|                             
;;         \/                      \/     \/                                             
;;
;;
;;
;;  This file is used to gather up the student's check-expects into a list.
;;
;;  This file will be require'd by the student's code. It's job
;;  is to collect the student's check-expects and make them available 
;;  via the get-check-expects function.
;;
;;  This file re-defines check-expect, check-within and check-error using
;;  define-syntax.
;;
;;  Provides:
;;     check-expect
;;     check-within
;;     check-error
;;     (add-a-testcase chexp)
;;         * chexp should be a list representing a check-expect function application
;;           (or a check-within or check-error function application)
;;     (get-all-testcases)
;;         * Returns all the testcases in a list, in the reverse order that the
;;           student wrote it in.
;;

(provide check-expect check-within check-error
         add-a-testcase get-all-testcases)

(define testcases empty)

;; Redefine check-expect, check-within, and check-error using define-syntax
(define-syntax check-expect
  (syntax-rules ()
    [(check-expect input ...)
     (add-a-testcase '(check-expect input ...))]))

(define-syntax check-within
  (syntax-rules ()
    [(check-within input ...)
     (add-a-testcase '(check-within input ...))]))

(define-syntax check-error
  (syntax-rules ()
    [(check-error input ...)
     (add-a-testcase '(check-error input ...))]))

;; add-a-testcase
(define (add-a-testcase chexpt)
  (set! testcases (cons chexpt testcases)))

;; get-all-testcases
(define (get-all-testcases) testcases)
