;; The first three lines of this file were inserted by DrRacket. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "htdp-intermediate-lambda-reader.ss" "lang")((modname skyscrapers) (read-case-sensitive #t) (teachpacks ()) (htdp-settings #(#t constructor repeating-decimal #f #t none #f ())))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Data definitions
;;
;; A Cell is a (union Nat[>0] '?)
;;
;; A Grid is a (ne-listof (ne-listof Cell))
;; A ClueSet is a (ne-listof Nat[>0])

(define-struct state (grid left right top bottom))
;; A State is a (make-state Grid ClueSet ClueSet ClueSet ClueSet)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; visibility-number: (listof Cell) -> Nat
;; Compute the visibility number of a list of Cells (loc).
;; The numeric entries should not have any duplicates.

;; Examples:

(check-expect (visibility-number empty) 0)
(check-expect (visibility-number '(4 5 6 9)) 4)
(check-expect (visibility-number '(4 ? 6 5 ? 9)) 3)

(define (visibility-number loc)
  (cond 
    [(empty? loc) 0]
    [else
     (local [;; vis-num-acc: (listof Nat [>0]) Nat -> Nat
             ;; Consumes a Nat [>0] called lon and
             ;; an accumulator holding the current
             ;; highest Nat in lon called high
             ;; and outputs the visibility number.
             ;; The numeric entries should not have any duplicates.
             
             (define (vis-num-acc lon high)
               (cond
                 [(empty? lon) 0]
                 [(< (first lon) high) (vis-num-acc (rest lon) high)]
                 [else (add1 (vis-num-acc (rest lon) (first lon)))]))]
       
       (vis-num-acc (filter number? loc) 0))]))

;; Tests:

(check-expect (visibility-number '(?)) 0)
(check-expect (visibility-number '(1)) 1)
(check-expect (visibility-number '(5 ? 2)) 1)
(check-expect (visibility-number '(1 ? ? 2)) 2)
(check-expect (visibility-number '(5 4 2)) 1)
(check-expect (visibility-number '(4 6 1 5)) 2)
(check-expect (visibility-number '(1 2 3 4 5 6)) 6)
(check-expect (visibility-number '(5 3 4 2 1 6)) 2)
(check-expect (visibility-number '(? 3 4 2 1 6)) 3)
(check-expect (visibility-number '(? 3 4 ? 1 6)) 3)
(check-expect (visibility-number '(? 3 ? ? 1 6)) 2)
(check-expect (visibility-number '(? 3 ? ? ? 6)) 2)
(check-expect (visibility-number '(? ? ? ? ? 6)) 1)
(check-expect (visibility-number '(? ? ? ? ? ?)) 0)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; feasible-visibility? : (listof Cell) Nat -> Boolean
;; Given a partially filled in row lov and a natural number vis 
;; that counts the number of towers that are supposed to be
;; visible, is it possible to fill in the blanks in lov so that
;; the desired visibility number is attained?  Assume that the
;; numeric entries in lov are in a contiguous block at the start
;; or end of the list.

;; Examples:

(check-expect (feasible-visibility? empty 0) true)
(check-expect (feasible-visibility? empty 2) false)
(check-expect (feasible-visibility? '(3 ? ? ?) 2) true)
(check-expect (feasible-visibility? '(2 1 ? ?) 4) false)

(define (feasible-visibility? lov vis)
  (cond
    [(and (empty? lov) (= 0 vis)) true]
    [(empty? lov) false]
    
    [else
     (local [;; n is the length of lov (as denoted in assignment package)
             (define n (length lov))
             
             ;; no-q is the same as lov, but without any unknown clues
             (define no-q (filter number? lov))
             
             ;; 1-to-n is a list of all Nats 1 to n
             ;; Lambda function's parameter x is the
             ;; current number, and we add 1 to it 
             ;; to get 1 ... n instead of 0 ... n-1
             (define 1-to-n (build-list n (lambda (x) (add1 x))))
             
             ;; s is the list of Nats between 1 and n
             ;; that do not appear in lov (in ascending order)
             ;; (as denoted in the assignment package)
             ;; Lambda function's parameter x is the current Nat.
             (define s (filter (lambda (x) (not (member? x lov))) 1-to-n))
             
             ;; rev-s is the reverse of s, that is,
             ;; it is the list of Nats between 1 and n
             ;; that do not appear in lov (in descending order)
             (define rev-s (reverse s))
             
             ;; fill-empty: (listof Nat) -> (listof Nat)
             ;; Consumes a list of Nat called lon (short for list of Nat)
             ;; and fills the empty clues in lov with lon, and outputs
             ;; that new list of nats.
             
             (define (fill-empty lon)
               (cond
                 [(number? (first lov)) (append no-q lon)]
                 [else (append lon no-q)]))
             
             ;; v-max is the list of Nats with the
             ;; highest possible visibility (as per assignment package)
             (define v-max (fill-empty s))
             
             ;; v-min is the list of Nats with the
             ;; lowest possible visibility (as per assignment package)
             (define v-min (fill-empty rev-s))]
       
       (and (<= vis (visibility-number v-max))
            (<= (visibility-number v-min) vis)))]))

;; Tests

(check-expect (feasible-visibility? '(?) 1) true)
(check-expect (feasible-visibility? '(2 ?) 1) true)
(check-expect (feasible-visibility? '(2 ?) 2) false)
(check-expect (feasible-visibility? '(? ? 2) 2) true)
(check-expect (feasible-visibility? '(? ? 3) 1) false)
(check-expect (feasible-visibility? '(? ? 3) 2) true)
(check-expect (feasible-visibility? '(? ? 3) 3) true)
(check-expect (feasible-visibility? '(? ? 3) 0) false)
(check-expect (feasible-visibility? '(? ? 3) 4) false)
(check-expect (feasible-visibility? '(3 2 ? ?) 2) true)
(check-expect (feasible-visibility? '(2 3 ? ?) 2) false)
(check-expect (feasible-visibility? '(2 3 1 ?) 2) false)
(check-expect (feasible-visibility? '(2 3 1 ?) 3) true)
(check-expect (feasible-visibility? '(? ? 2 3) 3) false)
(check-expect (feasible-visibility? '(? ? 2 3) 2) true)
(check-expect (feasible-visibility? '(? ? 2 3) 1) true)
(check-expect (feasible-visibility? '(? ? ? ?) 0) false)
(check-expect (feasible-visibility? '(? ? ? ?) 1) true)
(check-expect (feasible-visibility? '(? ? ? ?) 2) true)
(check-expect (feasible-visibility? '(? ? ? ?) 3) true)
(check-expect (feasible-visibility? '(? ? ? ?) 4) true)
(check-expect (feasible-visibility? '(? ? ? ?) 5) false)
(check-expect (feasible-visibility? '(? ? 5 ? ? ?) 543) false)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; valid-state? : State -> Boolean
;; Determine whether the passed-in State a-state can still be legally
;; continued to produce a solution for the entire puzzle.  Rule out
;; any State in which there are duplicate entries in one row or column
;; and any State in which a row or column violates its visibility
;; constraints.

;; Examples:

(check-expect 
 (valid-state? 
  (make-state '((? ? 3) (? ? ?) (? ? ?)) 
              '(1 2 1) '(2 1 1) '(1 1 1) '(2 2 2))) false)
(check-expect
 (valid-state?
  (make-state '((? ? 3) (? ? ?) (? ? ?))
              '(3 1 1) '(1 2 2) '(1 1 1) '(2 2 2))) true)
(check-expect
 (valid-state?
  (make-state '((? ? 3) (? ? ?) (? ? 3))
              '(3 1 1) '(1 2 2) '(1 1 1) '(2 2 2))) false)
(check-expect
 (valid-state?
  (make-state '((? ? ?) (? ? ?) (? ? ?))
              '(3 1 1) '(1 2 2) '(1 1 1) '(2 2 2))) true)

(define (valid-state? a-state)
  (local [;; check-all: Nat -> Noolean
          ;; Consumes a Nat called n which is initially the size of a-state's grid minus 1
          ;; and is used to see if we are done manipulating the grid
          ;; (it tells us what row/column to deal with)
          ;; and produces false if there are any duplicates in any
          ;; row or column or the visibility is impossible
          ;; at any end of a row or column,
          ;; otherwise produces true.
          
          (define (check-all n)
            (cond
              [(< n 0) true]
              [else
               (local [;; grid is the main grid we are manipulating
                       (define grid (state-grid a-state))
                       
                       ;; row is the nth row of grid 
                       (define row (list-ref grid n))
                       
                       ;; col is the nth column of grid 
                       ;; Lambda function's parameter row stores the current
                       ;; row in the grid, and x is a list of previous members
                       ;; of the current column which the implicit recursion found.
                       (define col (foldr (lambda (row x) (cons (list-ref row n) x)) empty grid))
                       
                       ;; dup-exist?: (listof Any) -> Boolean
                       ;; Consumes a listof Any called loa
                       ;; (short for list of any) and outputs
                       ;; true if any duplicate exists in loa,
                       ;; false if none exist.
                       
                       (define (dup-exist? loa)
                         (cond
                           [(empty? loa) false]
                           [(member? (first loa) (rest loa)) true]
                           [else (dup-exist? (rest loa))]))]
                 
                 (cond
                   [(or (dup-exist? (filter number? row)) (dup-exist? (filter number? col))
                        (not (feasible-visibility? row (list-ref (state-left a-state) n)))
                        (not (feasible-visibility? (reverse row) (list-ref (state-right a-state) n)))
                        (not (feasible-visibility? col (list-ref (state-top a-state) n)))
                        (not (feasible-visibility? (reverse col) (list-ref (state-bottom a-state) n)))) false]
                   [else (check-all (sub1 n))]))]))]
    
    (check-all (sub1 (length (state-grid a-state))))))

;; Tests:

(check-expect
 (valid-state?
  (make-state '((1 1 ?) (? ? ?) (? ? ?))
              '(3 1 1) '(1 2 2) '(1 1 1) '(2 2 2))) false)
(check-expect
 (valid-state?
  (make-state '((2 ? ?) (? ? ?) (? ? ?))
              '(1 1 1) '(1 1 1) '(1 1 1) '(2 2 2))) false)
(check-expect
 (valid-state?
  (make-state '((? ? 2) (? ? ?) (? ? 3))
              '(1 2 1) '(1 1 1) '(1 1 1) '(2 2 1))) false)
(check-expect 
 (valid-state? 
  (make-state '((? ? 3) (? ? ?) (? ? ?)) 
              '(1 2 1) '(2 1 1) '(1 1 1) '(2 2 2))) false)
(check-expect 
 (valid-state? 
  (make-state '((3 1 2) (2 3 1) (1 2 3))
              '(1 2 3) '(2 1 1) '(1 1 1) '(1 1 1))) false)
(check-expect 
 (valid-state? 
  (make-state '((3 1 2) (2 3 1) (1 2 3))
              '(1 2 3) '(2 1 1) '(1 1 2) '(1 1 1))) false)
(check-expect 
 (valid-state? 
  (make-state '((3 1 2) (2 3 1) (1 2 3))
              '(1 2 3) '(2 2 1) '(1 2 2) '(3 2 1))) true)
(check-expect 
 (valid-state? 
  (make-state '((3 ? ?) (? ? ?) (? ? 1))
              '(1 2 3) '(2 2 1) '(1 2 2) '(3 2 1))) false)
(check-expect 
 (valid-state? 
  (make-state '((3 ? ?) (? ? ?) (? ? ?))
              '(1 2 3) '(2 2 1) '(1 2 2) '(3 2 1))) true)
(check-expect 
 (valid-state? 
  (make-state '((? ? ?) (? ? ?) (? ? ?))
              '(1 2 3) '(2 2 1) '(1 2 2) '(3 2 1))) true)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; complete-state? : State -> Boolean
;; Determine whether the passed-in State a-state represents a solved
;; puzzle, i.e., one in which the grid has no unknowns left it in
;; (we assume that the puzzle is legal if this is the case).
;; Lambda function's parameter x simply contains each
;; successive row of the game-state.

;; Examples:

(check-expect
 (complete-state?
  (make-state '((? 1 2) (3 2 1) (1 2 3))
              '(1 1 1) '(1 1 1) '(1 1 1) '(1 1 1))) false)
(check-expect
 (complete-state?
  (make-state '((1 2) (2 1))
              '(2 1) '(1 2) '(2 1) '(1 2))) true)

(define (complete-state? a-state)
  (andmap (lambda (x) (andmap number? x)) (state-grid a-state)))

;; Tests:

(check-expect
 (complete-state?
  (make-state '((1 ?) (2 1))
              '(2 1) '(1 2) '(2 1) '(1 2))) false)
(check-expect
 (complete-state?
  (make-state '((1 ?) (2 ?))
              '(2 1) '(1 2) '(2 1) '(1 2))) false)
(check-expect
 (complete-state?
  (make-state '((1 ?) (? ?))
              '(2 1) '(1 2) '(2 1) '(1 2))) false)
(check-expect
 (complete-state?
  (make-state '((? ?) (? ?))
              '(2 1) '(1 2) '(2 1) '(1 2))) false)
(check-expect
 (complete-state?
  (make-state '((3 1 2) (3 2 1) (1 2 ?))
              '(1 1 1) '(1 1 1) '(1 1 1) '(1 1 1))) false)
(check-expect
 (complete-state?
  (make-state '((3 1 2) (3 ? 1) (1 2 ?))
              '(1 1 1) '(1 1 1) '(1 1 1) '(1 1 1))) false)
(check-expect
 (complete-state?
  (make-state '((3 1 2) (3 2 1) (1 2 3))
              '(1 1 1) '(1 1 1) '(1 1 1) '(1 1 1))) true)
(check-expect
 (complete-state?
  (make-state '((3 ? 2) (3 2 1) (1 2 3))
              '(1 1 1) '(1 1 1) '(1 1 1) '(1 1 1))) false)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; neighbours : State -> (listof State)
;; Given a current State a-state, produce a list of all legal successor
;; states, according to the constraints laid out in the assignment (that
;; is, fill in the first non-numeric entry in the grid with each of the
;; legal numbers that can live in that spot.

;; Examples:
(check-expect
 (neighbours
  (make-state '((?)) '(1) '(1) '(1) '(1)))
 (list (make-state '((1)) '(1) '(1) '(1) '(1))))
(check-expect
 (neighbours
  (make-state
   (make-list 4 (make-list 4 '?))
   '(1 3 2 2) '(3 2 2 1) '(1 3 2 3) '(2 2 3 1)))
 (list (make-state '((4 ? ? ?) (? ? ? ?) (? ? ? ?) (? ? ? ?))
                   '(1 3 2 2) '(3 2 2 1) '(1 3 2 3) '(2 2 3 1))))

(define (neighbours a-state)
  (local [;; fill-first-empty: Nat -> State
          ;; Comsumes a Nat called n and fills
          ;; the first empty space in the grid
          ;; of a-state with n, and then outputs
          ;; that new state.
          
          (define (fill-first-empty n)
            (local [;; fill-first-empty/list: (listof Cell) (listof Cell) -> (listof Cell)
                    ;; Consumes a (listof Cell) called loc
                    ;; (short for list of Cell) and
                    ;; and another listof Cell called loc-acc
                    ;; which accumulates a list of all Cells 
                    ;; before finding the first instance
                    ;; of an empty space and outputs
                    ;; a (listof Cell) with the first instance of
                    ;; an empty space replaced with the Nat n.
                    
                    (define (fill-first-empty/list loc loc-acc)
                      (cond
                        [(empty? loc) empty]
                        [(symbol? (first loc)) (append loc-acc (list n) (rest loc))]
                        [else
                         (fill-first-empty/list (rest loc) (append loc-acc (list (first loc))))]))
                    
                    ;; fill-first-empty/grid: Grid -> Grid
                    ;; Consumes a Grid called grid and
                    ;; replaces the first empty cell with
                    ;; the Nat n. Assumes there is at least one empty.
                    
                    (define (fill-first-empty/grid grid)
                      (cond
                        [(empty? grid) empty]
                        [(not (member? '? (first grid)))
                         (append (list (first grid)) (fill-first-empty/grid (rest grid)))]
                        [else
                         (append (list (fill-first-empty/list (first grid) empty)) (rest grid))]))]
              
              (make-state (fill-first-empty/grid (state-grid a-state))
                          (state-left a-state) (state-right a-state) 
                          (state-top a-state) (state-bottom a-state))))
          
          ;; n is the length of the grid of a-state 
          ;; (as denoted in the assignment package)
          (define n (length (state-grid a-state)))
          
          ;; 1-to-n is a list of all Nats 1 to n 
          ;; (which is the length of the grid of a-state)
          ;; Lambda function's parameter x is the
          ;; current number, and we add 1 to it 
          ;; to get 1 ... n instead of 0 ... n-1
          (define 1-to-n (build-list n (lambda (x) (add1 x))))
          
          ;; candidate-states are all possibilities of filling the
          ;; first empty space with all numbers 1 through n.
          ;; Its lambda function used in map simply
          (define candidate-states (map (lambda (x) (fill-first-empty x)) 1-to-n))]
    
    (filter valid-state? candidate-states)))

;; Tests:

(check-expect
 (neighbours
  (make-state
   (make-list 3 (make-list 3 '?))
   '(2 3 2) '(3 2 2) '(1 3 2) '(2 2 3))) empty)

(check-expect
 (neighbours
  (make-state
   (make-list 3 (make-list 3 '?))
   '(1 3 2) '(3 2 2) '(1 3 2) '(2 2 3)))
 (list (make-state '((3 ? ?) (? ? ?) (? ? ?))
                   '(1 3 2) '(3 2 2) '(1 3 2) '(2 2 3))))
(check-expect
 (neighbours
  (make-state
   (make-list 3 (make-list 3 '?))
   '(1 3 2) '(3 2 2) '(1 3 2) '(2 2 3)))
 (list (make-state '((3 ? ?) (? ? ?) (? ? ?))
                   '(1 3 2) '(3 2 2) '(1 3 2) '(2 2 3))))
(check-expect
 (neighbours
  (make-state
   (make-list 3 (make-list 3 '?))
   '(1 3 2) '(3 1 2) '(1 3 2) '(3 2 2)))
 (list (make-state '((3 ? ?) (? ? ?) (? ? ?))
                   '(1 3 2) '(3 1 2) '(1 3 2) '(3 2 2))))
(check-expect
 (neighbours
  (make-state
   (make-list 3 (make-list 3 '?))
   '(2 3 2) '(3 1 2) '(1 3 2) '(3 2 2))) empty)

(check-expect
 (neighbours
  (make-state '((3 1 2) (? 2 1) (1 2 3))
              '(1 1 1) '(1 1 1) '(1 1 1) '(1 1 1))) empty)

(check-expect
 (neighbours
  (make-state '((3 ? ?) (? ? ?) (? ? ?))
              '(1 2 3) '(2 1 3) '(1 2 3) '(2 1 3)))
 (list (make-state '((3 1 ?) (? ? ?) (? ? ?))
                   '(1 2 3) '(2 1 3) '(1 2 3) '(2 1 3))))
(check-expect
 (neighbours
  (make-state '((3 1 ?) (? ? ?) (? ? ?))
              '(1 2 3) '(2 1 3) '(1 2 3) '(2 1 3))) empty)

(check-expect (neighbours puzzle-1)
              (list (make-state (list (list 4 '? '? '?) 
                                      (list '? '? '? '?) 
                                      (list '? '? '? '?) 
                                      (list '? '? '? '?)) 
                                (list 1 2 2 3) (list 3 3 1 2) 
                                (list 1 2 3 2) (list 3 2 1 2))))

(check-expect (neighbours puzzle-2)
              (list (make-state (list (list 4 '? '? '?) 
                                      (list '? '? '? '?) 
                                      (list '? '? '? '?) 
                                      (list '? '? '? '?)) 
                                (list 1 2 3 2) (list 3 3 2 1) 
                                (list 1 2 3 3) (list 2 3 2 1))))

(check-expect (neighbours (make-state (list (list 4 '? '? '?) 
                                            (list '? '? '? '?) 
                                            (list '? '? '? '?) 
                                            (list '? '? '? '?)) 
                                      (list 1 2 3 2) (list 3 3 2 1) 
                                      (list 1 2 3 3) (list 2 3 2 1)))
              (list
               (make-state
                (list
                 (list 4 1 '? '?)
                 (list '? '? '? '?)
                 (list '? '? '? '?)
                 (list '? '? '? '?))
                (list 1 2 3 2)
                (list 3 3 2 1)
                (list 1 2 3 3)
                (list 2 3 2 1))
               (make-state
                (list
                 (list 4 2 '? '?)
                 (list '? '? '? '?)
                 (list '? '? '? '?)
                 (list '? '? '? '?))
                (list 1 2 3 2)
                (list 3 3 2 1)
                (list 1 2 3 3)
                (list 2 3 2 1))
               (make-state
                (list
                 (list 4 3 '? '?)
                 (list '? '? '? '?)
                 (list '? '? '? '?)
                 (list '? '? '? '?))
                (list 1 2 3 2)
                (list 3 3 2 1)
                (list 1 2 3 3)
                (list 2 3 2 1))))

(check-expect (neighbours (make-state (list (list 4 3 1 2) 
                                            (list 2 4 3 1) 
                                            (list 1 2 4 '?) 
                                            (list '? '? '? '?)) 
                                      (list 1 2 3 2) (list 3 3 2 1) 
                                      (list 1 2 3 3) (list 2 3 2 1)))
              (list (make-state (list (list 4 3 1 2) 
                                      (list 2 4 3 1) 
                                      (list 1 2 4 3) 
                                      (list '? '? '? '?)) 
                                (list 1 2 3 2) (list 3 3 2 1) 
                                (list 1 2 3 3) (list 2 3 2 1))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; search: (X -> Boolean) (X -> (listof X)) X -> (union X false)
;; Solve a puzzle, given a current State.  Use depth-first search
;; with backtracking, assuming that the search space is acyclic. 
;; The parameters are:
;;  * at-end? -- a predicate that determines whether the given
;;      state is an end state of the search (i.e., a solution).
;;  * neighbours -- a function that maps a current state to a list
;;      of neighbouring states.
;;  * a-state -- the initial state.
;;
;; DO NOT MODIFY THIS FUNCTION.

(define (search at-end? neighbours a-state)
  (local
    [;; find-route: X -> (union X false)
     ;; Search outward from this configuration to see if there's a path
     ;; to a solution.
     (define (find-route a-state)
       (cond
         [(at-end? a-state) a-state]
         [else (find-route/list (neighbours a-state))]))
     
     ;; find-route/list: (listof X) -> (union X false)
     ;; Search outward from every configuration in the passed-in list of
     ;; configurations.  If any one of them leads to a solution, stop and
     ;; produce that solution.  Produce false if you run out of options.
     (define (find-route/list lostate)
       (cond
         [(empty? lostate) false]
         [else
          (local
            [(define cur (find-route (first lostate)))]
            (cond
              [(not (boolean? cur)) cur]
              [else (find-route/list (rest lostate))]))]))]
    (find-route a-state)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Sample puzzles for testing.  You can easily create more yourself by 
;; transcribing them from examples generated using the software mentioned
;; in the assignment.
;;

;; The puzzle that appears at the start of the assignment handout.
(define assignment-puzzle
  (make-state
   (make-list 4 (make-list 4 '?))
   '(1 3 2 2)
   '(3 2 2 1)
   '(1 3 2 3)
   '(2 2 3 1)))
(define assignment-puzzle-solution
  (make-state
   '((4 1 3 2) (2 3 4 1) (1 4 2 3) (3 2 1 4))
   (state-left assignment-puzzle)
   (state-right assignment-puzzle)
   (state-top assignment-puzzle)
   (state-bottom assignment-puzzle)))

(define puzzle-1 
  (make-state
   (make-list 4 (make-list 4 '?))
   '(1 2 2 3)
   '(3 3 1 2)
   '(1 2 3 2)
   '(3 2 1 2)))
(define puzzle-1-solution 
  (make-state
   '((4 3 1 2) (2 4 3 1) (3 1 2 4) (1 2 4 3))
   (state-left puzzle-1)
   (state-right puzzle-1)
   (state-top puzzle-1)
   (state-bottom puzzle-1)))

(define puzzle-2 
  (make-state
   (make-list 4 (make-list 4 '?))
   '(1 2 3 2)
   '(3 3 2 1)
   '(1 2 3 3)
   '(2 3 2 1)))
(define puzzle-2-solution 
  (make-state
   '((4 3 1 2) 
     (2 4 3 1) 
     (1 2 4 3) 
     (3 1 2 4))
   (state-left puzzle-2)
   (state-right puzzle-2)
   (state-top puzzle-2)
   (state-bottom puzzle-2)))

(define puzzle-3 
  (make-state
   (make-list 5 (make-list 5 '?))
   '(2 3 2 1 3)
   '(2 1 2 3 2)
   '(4 1 2 3 2) 
   '(2 5 2 1 2)))
(define puzzle-3-solution 
  (make-state
   '((1 5 2 3 4) 
     (3 4 1 2 5)
     (4 3 5 1 2)
     (5 2 3 4 1)
     (2 1 4 5 3))
   (state-left puzzle-3)
   (state-right puzzle-3)
   (state-top puzzle-3)
   (state-bottom puzzle-3)))

(define puzzle-4 
  (make-state
   (make-list 5 (make-list 5 '?))
   '(2 2 1 2 4)
   '(3 2 5 3 1)
   '(3 1 3 2 3) 
   '(3 3 2 3 1)))
(define puzzle-4-solution 
  (make-state
   '((1 5 2 4 3) 
     (3 2 1 5 4)
     (5 4 3 2 1)
     (4 1 5 3 2)
     (2 3 4 1 5))
   (state-left puzzle-4)
   (state-right puzzle-4)
   (state-top puzzle-4)
   (state-bottom puzzle-4)))

(define bonus-puzzle
  (make-state
   '((? ? ? ? ? ?)
     (? ? ? ? ? ?)
     (? ? ? ? ? ?)
     (? ? ? ? ? ?)
     (? ? ? ? ? ?)
     (? ? ? 2 ? ?))
   '(2 4 ? ? ? ?)
   '(4 ? 4 ? 2 2)
   '(? 2 ? ? ? 4)
   '(? ? ? ? 4 ?)))

;; Tests:

(check-expect (search complete-state? neighbours assignment-puzzle) assignment-puzzle-solution)
(check-expect (search complete-state? neighbours puzzle-1) puzzle-1-solution)
(check-expect (search complete-state? neighbours puzzle-2) puzzle-2-solution)
(check-expect (search complete-state? neighbours puzzle-3) puzzle-3-solution)
(check-expect (search complete-state? neighbours puzzle-4) puzzle-4-solution)
